@@ -21,6 +21,7 @@
 
 #include "akdebug.h"
 #include "akdbus.h"
+#include "akonadi.h"
 #include "connection.h"
 #include "handler.h"
 #include "handlerhelper.h"
@@ -38,6 +39,7 @@
 #include "utils.h"
 #include "intervalcheck.h"
 #include "agentmanagerinterface.h"
+#include "dbusconnectionpool.h"
 
 #include <QtCore/QLocale>
 #include <QtCore/QStringList>
@@ -110,7 +112,7 @@ QSqlQuery FetchHelper::buildPartQuery( const QVector<QByteArray> &partList, bool
       partQuery.addCondition( cond );
     }
 
-    ItemQueryHelper::scopeToQuery( mScope, mConnection, partQuery );
+    ItemQueryHelper::scopeToQuery( mScope, mConnection->context(), partQuery );
 
     if ( !partQuery.exec() ) {
       throw HandlerException( "Unable to list item parts" );
@@ -155,7 +157,7 @@ QSqlQuery FetchHelper::buildItemQuery()
   itemQuery.addSortColumn( PimItem::idFullColumnName(), Query::Descending );
 
   if ( mScope.scope() != Scope::Invalid ) {
-    ItemQueryHelper::scopeToQuery( mScope, mConnection, itemQuery );
+    ItemQueryHelper::scopeToQuery( mScope, mConnection->context(), itemQuery );
   }
 
   if ( mFetchScope.changedSince().isValid() ) {
@@ -185,7 +187,7 @@ QSqlQuery FetchHelper::buildFlagQuery()
                      Flag::idFullColumnName(), PimItemFlagRelation::rightFullColumnName() );
   flagQuery.addColumn( PimItem::idFullColumnName() );
   flagQuery.addColumn( Flag::nameFullColumnName() );
-  ItemQueryHelper::scopeToQuery( mScope, mConnection, flagQuery );
+  ItemQueryHelper::scopeToQuery( mScope, mConnection->context(), flagQuery );
   flagQuery.addSortColumn( PimItem::idFullColumnName(), Query::Descending );
 
   if ( !flagQuery.exec() ) {
@@ -211,7 +213,7 @@ QSqlQuery FetchHelper::buildTagQuery()
                      Tag::idFullColumnName(), PimItemTagRelation::rightFullColumnName() );
   tagQuery.addColumn( PimItem::idFullColumnName() );
   tagQuery.addColumn( Tag::idFullColumnName() );
-  ItemQueryHelper::scopeToQuery( mScope, mConnection, tagQuery );
+  ItemQueryHelper::scopeToQuery( mScope, mConnection->context(), tagQuery );
   tagQuery.addSortColumn( PimItem::idFullColumnName(), Query::Descending );
 
   if ( !tagQuery.exec() ) {
@@ -223,6 +225,32 @@ QSqlQuery FetchHelper::buildTagQuery()
   return tagQuery.query();
 }
 
+enum VRefQueryColumns {
+  VRefQueryCollectionIdColumn,
+  VRefQueryItemIdColumn
+};
+
+QSqlQuery FetchHelper::buildVRefQuery()
+{
+  QueryBuilder vRefQuery( PimItem::tableName() );
+  vRefQuery.addJoin( QueryBuilder::LeftJoin, CollectionPimItemRelation::tableName(),
+                     CollectionPimItemRelation::rightFullColumnName(),
+                     PimItem::idFullColumnName() );
+  vRefQuery.addColumn( CollectionPimItemRelation::leftFullColumnName() );
+  vRefQuery.addColumn( CollectionPimItemRelation::rightFullColumnName() );
+  ItemQueryHelper::scopeToQuery( mScope, mConnection->context(), vRefQuery );
+  vRefQuery.addSortColumn( PimItem::idFullColumnName(), Query::Descending );
+
+  if (!vRefQuery.exec() ) {
+    throw HandlerException( "Unable to retrieve virtual references" );
+  }
+
+  vRefQuery.query().next();
+
+  return vRefQuery.query();
+}
+
+
 bool FetchHelper::isScopeLocal( const Scope &scope )
 {
   // Get list of all resources that own all items in the scope
@@ -231,9 +259,9 @@ bool FetchHelper::isScopeLocal( const Scope &scope )
   qb.addColumn( Resource::nameFullColumnName() );
   qb.addJoin( QueryBuilder::LeftJoin, Collection::tableName(), PimItem::collectionIdFullColumnName(), Collection::idFullColumnName() );
   qb.addJoin( QueryBuilder::LeftJoin, Resource::tableName(), Collection::resourceIdFullColumnName(), Resource::idFullColumnName() );
-  ItemQueryHelper::scopeToQuery( scope, mConnection, qb );
-  if ( mConnection->resourceContext().isValid() ) {
-    qb.addValueCondition( Resource::nameFullColumnName(), Query::NotEquals, mConnection->resourceContext().name() );
+  ItemQueryHelper::scopeToQuery( scope, mConnection->context(), qb );
+  if ( mConnection->context()->resource().isValid() ) {
+    qb.addValueCondition( Resource::nameFullColumnName(), Query::NotEquals, mConnection->context()->resource().name() );
   }
 
   if ( !qb.exec() ) {
@@ -252,16 +280,12 @@ bool FetchHelper::isScopeLocal( const Scope &scope )
 
   query.next();
   const QString resourceName = query.value( 0 ).toString();
-  // Workaround for QDBusConnectionPrivate not being thread-safe in Qt 4, fixed in Qt 5.2
-  // TODO: Remove in KF5
-  const QDBusConnection connection = QDBusConnection::connectToBus( QDBusConnection::SessionBus,
-                                                                    QString::fromLatin1( mConnection->sessionId() ) );
+
   org::freedesktop::Akonadi::AgentManager manager( AkDBus::serviceName( AkDBus::Control ),
                                                    QLatin1String( "/AgentManager" ),
-                                                   connection );
+                                                   DBusConnectionPool::threadConnection() );
   const QString typeIdentifier = manager.agentInstanceType( resourceName );
   const QVariantMap properties = manager.agentCustomProperties( typeIdentifier );
-  QDBusConnection::disconnectFromBus( QString::fromLatin1( mConnection->sessionId() ) );
   return properties.value( QLatin1String( "HasLocalStorage" ), false ).toBool();
 }
 
@@ -286,14 +310,14 @@ bool FetchHelper::fetchItems( const QByteArray &responseIdentifier )
     retriever.setRetrieveFullPayload( mFetchScope.fullPayload() );
     retriever.setChangedSince( mFetchScope.changedSince() );
     if ( !retriever.exec() && !mFetchScope.ignoreErrors() ) { // There we go, retrieve the missing parts from the resource.
-      if ( mConnection->resourceContext().isValid() ) {
+      if ( mConnection->context()->resource().isValid() ) {
         throw HandlerException( QString::fromLatin1( "Unable to fetch item from backend (collection %1, resource %2) : %3" )
-                .arg( mConnection->selectedCollectionId() )
-                .arg( mConnection->resourceContext().id() )
+                .arg( mConnection->context()->collectionId() )
+                .arg( mConnection->context()->resource().id() )
                 .arg( QString::fromLatin1( retriever.lastError() ) ) );
       } else {
         throw HandlerException( QString::fromLatin1( "Unable to fetch item from backend (collection %1) : %2" )
-                .arg( mConnection->selectedCollectionId() )
+                .arg( mConnection->context()->collectionId() )
                 .arg( QString::fromLatin1( retriever.lastError() ) ) );
       }
     }
@@ -336,6 +360,11 @@ bool FetchHelper::fetchItems( const QByteArray &responseIdentifier )
     tagQuery = buildTagQuery();
   }
 
+  QSqlQuery vRefQuery;
+  if ( mFetchScope.virtualReferencesRequested() ) {
+    vRefQuery = buildVRefQuery();
+  }
+
   // build responses
   Response response;
   response.setUntagged();
@@ -370,7 +399,7 @@ bool FetchHelper::fetchItems( const QByteArray &responseIdentifier )
       }
     }
     if ( mFetchScope.gidRequested() ) {
-      const QByteArray gid = Utils::variantToByteArray( itemQuery.value( ItemQueryPimItemGidColumn ) );
+      const QByteArray gid = Utils::variantToByteArray( extractQueryResult( itemQuery, ItemQueryPimItemGidColumn ) );
       if ( !gid.isEmpty() ) {
         attributes.append( AKONADI_PARAM_GID " " + ImapParser::quote( gid ) );
       }
@@ -411,6 +440,25 @@ bool FetchHelper::fetchItems( const QByteArray &responseIdentifier )
       }
     }
 
+    if ( mFetchScope.virtualReferencesRequested() ) {
+      ImapSet cols;
+      while ( vRefQuery.isValid() ) {
+          const qint64 id = vRefQuery.value( VRefQueryItemIdColumn ).toLongLong();
+          if ( id > pimItemId ) {
+            vRefQuery.next();
+            continue;
+          } else if ( id < pimItemId ) {
+            break;
+          }
+          const qint64 collectionId = vRefQuery.value( VRefQueryCollectionIdColumn ).toLongLong();
+          cols.add( QVector<qint64>() << collectionId );
+          vRefQuery.next();
+      }
+      if ( !cols.isEmpty() ) {
+        attributes.append( AKONADI_PARAM_VIRTREF " " + cols.toImapSequenceSet() );
+      }
+    }
+
     if ( mFetchScope.ancestorDepth() > 0 ) {
       attributes.append( HandlerHelper::ancestorsToByteArray( mFetchScope.ancestorDepth(), ancestorsForItem( parentCollectionId ) ) );
     }
@@ -520,7 +568,7 @@ void FetchHelper::updateItemAccessTime()
   Transaction transaction( mConnection->storageBackend() );
   QueryBuilder qb( PimItem::tableName(), QueryBuilder::Update );
   qb.setColumnValue( PimItem::atimeColumn(), QDateTime::currentDateTime() );
-  ItemQueryHelper::scopeToQuery( mScope, mConnection, qb );
+  ItemQueryHelper::scopeToQuery( mScope, mConnection->context(), qb );
 
   if ( !qb.exec() ) {
     qWarning() << "Unable to update item access time";
@@ -531,11 +579,11 @@ void FetchHelper::updateItemAccessTime()
 
 void FetchHelper::triggerOnDemandFetch()
 {
-  if ( mScope.scope() != Scope::None || mConnection->selectedCollectionId() <= 0 || mFetchScope.cacheOnly() ) {
+  if ( mScope.scope() != Scope::None || mConnection->context()->collectionId() <= 0 || mFetchScope.cacheOnly() ) {
     return;
   }
 
-  Collection collection = mConnection->selectedCollection();
+  Collection collection = mConnection->context()->collection();
 
   // HACK: don't trigger on-demand syncing if the resource is the one triggering it
   if ( mConnection->sessionId() == collection.resource().name().toLatin1() ) {
@@ -548,7 +596,9 @@ void FetchHelper::triggerOnDemandFetch()
     return;
   }
 
-  IntervalCheck::self()->requestCollectionSync( collection );
+  if ( AkonadiServer::instance()->intervalChecker() ) {
+    AkonadiServer::instance()->intervalChecker()->requestCollectionSync( collection );
+  }
 }
 
 QStack<Collection> FetchHelper::ancestorsForItem( Collection::Id parentColId )
