@@ -61,8 +61,17 @@ using namespace Akonadi::Server;
 static QMutex sTransactionMutex;
 bool DataStore::s_hasForeignKeyConstraints = false;
 
-#define TRANSACTION_MUTEX_LOCK if ( DbType::type( m_database ) == DbType::Sqlite ) sTransactionMutex.lock()
-#define TRANSACTION_MUTEX_UNLOCK if ( DbType::type( m_database ) == DbType::Sqlite ) sTransactionMutex.unlock()
+QThreadStorage<DataStore*> DataStore::sInstances;
+
+#define TRANSACTION_MUTEX_LOCK if ( DbType::isSystemSQLite( m_database ) ) sTransactionMutex.lock()
+#define TRANSACTION_MUTEX_UNLOCK if ( DbType::isSystemSQLite( m_database ) ) sTransactionMutex.unlock()
+
+#define setBoolPtr(ptr, val) \
+{ \
+  if ((ptr)) { \
+      *(ptr) = (val); \
+  } \
+}
 
 /***************************************************************************
  *   DataStore                                                           *
@@ -71,12 +80,11 @@ DataStore::DataStore()
   : QObject()
   , m_dbOpened( false )
   , m_transactionLevel( 0 )
-  , mNotificationCollector( new NotificationCollector( this ) )
+  , mNotificationCollector( 0 )
   , m_keepAliveTimer( 0 )
 {
   open();
-
-  NotificationManager::self()->connectNotificationCollector( mNotificationCollector );
+  notificationCollector();
 
   if ( DbConfig::configuredDatabase()->driverName() == QLatin1String( "QMYSQL" ) ) {
     // Send a dummy query to MySQL every 1 hour to keep the connection alive,
@@ -156,9 +164,13 @@ bool DataStore::init()
   }
   s_hasForeignKeyConstraints = initializer->hasForeignKeyConstraints();
 
-  DbUpdater updater( m_database, QLatin1String( ":dbupdate.xml" ) );
-  if ( !updater.run() ) {
-    return false;
+  if ( QFile::exists( QLatin1String( ":dbupdate.xml" ) ) ) {
+    DbUpdater updater( m_database, QLatin1String( ":dbupdate.xml" ) );
+    if ( !updater.run() ) {
+      return false;
+    }
+  } else {
+    qWarning() << "Warning: dbupdate.xml not found, skipping updates";
   }
 
   if ( !initializer->updateIndexesAndConstraints() ) {
@@ -175,19 +187,28 @@ bool DataStore::init()
   return true;
 }
 
-QThreadStorage<DataStore *> instances;
+NotificationCollector *DataStore::notificationCollector()
+{
+  if ( mNotificationCollector == 0 ) {
+    mNotificationCollector = new NotificationCollector( this );
+    NotificationManager::self()->connectNotificationCollector( notificationCollector() );
+  }
+
+  return mNotificationCollector;
+}
 
 DataStore *DataStore::self()
 {
-  if ( !instances.hasLocalData() ) {
-    instances.setLocalData( new DataStore() );
+  if ( !sInstances.hasLocalData() ) {
+    sInstances.setLocalData( new DataStore() );
   }
-  return instances.localData();
+  return sInstances.localData();
 }
 
 /* --- ItemFlags ----------------------------------------------------- */
 
-bool DataStore::setItemsFlags( const PimItem::List &items, const QVector<Flag> &flags )
+bool DataStore::setItemsFlags( const PimItem::List &items, const QVector<Flag> &flags,
+                               bool *flagsChanged, bool silent )
 {
   QSet<QByteArray> removedFlags;
   QSet<QByteArray> addedFlags;
@@ -195,6 +216,8 @@ bool DataStore::setItemsFlags( const PimItem::List &items, const QVector<Flag> &
   QVariantList insFlags;
   Query::Condition delConds( Query::Or );
 
+  setBoolPtr( flagsChanged, false );
+
   Q_FOREACH ( const PimItem &item, items ) {
     Q_FOREACH ( const Flag &flag, item.flags() ) {
       if ( !flags.contains( flag ) ) {
@@ -233,12 +256,11 @@ bool DataStore::setItemsFlags( const PimItem::List &items, const QVector<Flag> &
     }
   }
 
-  if ( addedFlags.empty() && removedFlags.empty() ) {
-    // no changes done, notification not needed
-    return true;
+  if ( !silent && ( !addedFlags.isEmpty() || !removedFlags.isEmpty() ) ) {
+    mNotificationCollector->itemsFlagsChanged( items, addedFlags, removedFlags );
   }
 
-  mNotificationCollector->itemsFlagsChanged( items, addedFlags, removedFlags );
+  setBoolPtr( flagsChanged, ( addedFlags != removedFlags ) );
 
   return true;
 }
@@ -282,7 +304,7 @@ bool DataStore::doAppendItemsFlag( const PimItem::List &items, const Flag &flag,
 }
 
 bool DataStore::appendItemsFlags( const PimItem::List &items, const QVector<Flag> &flags,
-                                  bool &flagsChanged, bool checkIfExists,
+                                  bool *flagsChanged, bool checkIfExists,
                                   const Collection &col, bool silent )
 {
   QSet<QByteArray> added;
@@ -292,7 +314,8 @@ bool DataStore::appendItemsFlags( const PimItem::List &items, const QVector<Flag
     itemsIds.append( item.id() );
   }
 
-  flagsChanged = false;
+  setBoolPtr( flagsChanged, false );
+
   Q_FOREACH ( const Flag &flag, flags ) {
     QSet<PimItem::Id> existing;
     if ( checkIfExists ) {
@@ -309,15 +332,23 @@ bool DataStore::appendItemsFlags( const PimItem::List &items, const QVector<Flag
       }
 
       QSqlQuery query = qb.query();
-      if ( query.size() == items.count() ) {
-        continue;
-      }
+      if ( query.driver()->hasFeature( QSqlDriver::QuerySize ) ) {
+        //The query size feature is not suppoerted by the sqllite driver
+        if ( query.size() == items.count() ) {
+          continue;
+        }
+        setBoolPtr( flagsChanged, true );
 
-      flagsChanged = true;
+      }
 
       while ( query.next() ) {
         existing << query.value( 0 ).value<PimItem::Id>();
       }
+      if ( !query.driver()->hasFeature( QSqlDriver::QuerySize ) ) {
+        if ( existing.size() != items.count() ) {
+          setBoolPtr( flagsChanged, true );
+        }
+      }
     }
 
     if ( !doAppendItemsFlag( items, flag, existing, col, silent ) ) {
@@ -328,11 +359,15 @@ bool DataStore::appendItemsFlags( const PimItem::List &items, const QVector<Flag
   return true;
 }
 
-bool DataStore::removeItemsFlags( const PimItem::List &items, const QVector<Flag> &flags )
+bool DataStore::removeItemsFlags( const PimItem::List &items, const QVector<Flag> &flags,
+                                  bool *flagsChanged, bool silent )
 {
   QSet<QByteArray> removedFlags;
   QVariantList itemsIds;
   QVariantList flagsIds;
+
+  setBoolPtr( flagsChanged, false );
+
   Q_FOREACH ( const PimItem &item, items ) {
     itemsIds << item.id();
     for ( int i = 0; i < flags.count(); ++i ) {
@@ -354,13 +389,20 @@ bool DataStore::removeItemsFlags( const PimItem::List &items, const QVector<Flag
     return false;
   }
 
-  mNotificationCollector->itemsFlagsChanged( items, QSet<QByteArray>(), removedFlags );
+  if ( qb.query().numRowsAffected() != 0 ) {
+    setBoolPtr( flagsChanged, true );
+    if ( !silent ) {
+      mNotificationCollector->itemsFlagsChanged( items, QSet<QByteArray>(), removedFlags );
+    }
+  }
+
   return true;
 }
 
 /* --- ItemTags ----------------------------------------------------- */
 
-bool DataStore::setItemsTags( const PimItem::List &items, const Tag::List &tags )
+bool DataStore::setItemsTags( const PimItem::List &items, const Tag::List &tags,
+                              bool *tagsChanged, bool silent )
 {
   QSet<qint64> removedTags;
   QSet<qint64> addedTags;
@@ -368,6 +410,8 @@ bool DataStore::setItemsTags( const PimItem::List &items, const Tag::List &tags
   QVariantList insTags;
   Query::Condition delConds( Query::Or );
 
+  setBoolPtr( tagsChanged, false );
+
   Q_FOREACH ( const PimItem &item, items ) {
     Q_FOREACH ( const Tag &tag, item.tags() ) {
       if ( !tags.contains( tag ) ) {
@@ -408,18 +452,18 @@ bool DataStore::setItemsTags( const PimItem::List &items, const Tag::List &tags
     }
   }
 
-  if ( addedTags.empty() && removedTags.empty() ) {
-    // no changes done, notification not needed
-    return true;
+  if ( !silent && ( !addedTags.empty() || !removedTags.empty() ) ) {
+    mNotificationCollector->itemsTagsChanged( items, addedTags, removedTags );
   }
 
-  mNotificationCollector->itemsTagsChanged( items, addedTags, removedTags );
+  setBoolPtr( tagsChanged, ( addedTags != removedTags ) );
 
   return true;
 }
 
 bool DataStore::doAppendItemsTag( const PimItem::List &items, const Tag &tag,
-                                   const QSet<Entity::Id> &existing, const Collection &col )
+                                  const QSet<Entity::Id> &existing, const Collection &col,
+                                  bool silent )
 {
   QVariantList tagIds;
   QVariantList appendIds;
@@ -447,15 +491,17 @@ bool DataStore::doAppendItemsTag( const PimItem::List &items, const Tag &tag,
     return false;
   }
 
-  mNotificationCollector->itemsTagsChanged( appendItems, QSet<qint64>() << tag.id(),
-                                             QSet<qint64>(), col );
+  if ( !silent ) {
+    mNotificationCollector->itemsTagsChanged( appendItems, QSet<qint64>() << tag.id(),
+                                               QSet<qint64>(), col );
+  }
 
   return true;
 }
 
 bool DataStore::appendItemsTags( const PimItem::List &items, const Tag::List &tags,
-                                  bool &tagsChanged, bool checkIfExists,
-                                  const Collection &col )
+                                  bool *tagsChanged, bool checkIfExists,
+                                  const Collection &col, bool silent )
 {
   QSet<QByteArray> added;
 
@@ -464,7 +510,8 @@ bool DataStore::appendItemsTags( const PimItem::List &items, const Tag::List &ta
     itemsIds.append( item.id() );
   }
 
-  tagsChanged = false;
+  setBoolPtr( tagsChanged, false );
+
   Q_FOREACH ( const Tag &tag, tags ) {
     QSet<PimItem::Id> existing;
     if ( checkIfExists ) {
@@ -485,13 +532,13 @@ bool DataStore::appendItemsTags( const PimItem::List &items, const Tag::List &ta
         continue;
       }
 
-      tagsChanged = true;
+      setBoolPtr( tagsChanged, true );
 
       while ( query.next() ) {
         existing << query.value( 0 ).value<PimItem::Id>();
       } }
 
-    if ( !doAppendItemsTag( items, tag, existing, col ) ) {
+    if ( !doAppendItemsTag( items, tag, existing, col, silent ) ) {
       return false;
     }
   }
@@ -499,11 +546,15 @@ bool DataStore::appendItemsTags( const PimItem::List &items, const Tag::List &ta
   return true;
 }
 
-bool DataStore::removeItemsTags( const PimItem::List &items, const Tag::List &tags )
+bool DataStore::removeItemsTags( const PimItem::List &items, const Tag::List &tags,
+                                 bool *tagsChanged, bool silent )
 {
   QSet<qint64> removedTags;
   QVariantList itemsIds;
   QVariantList tagsIds;
+
+  setBoolPtr( tagsChanged, false );
+
   Q_FOREACH ( const PimItem &item, items ) {
     itemsIds << item.id();
     for ( int i = 0; i < tags.count(); ++i ) {
@@ -525,7 +576,13 @@ bool DataStore::removeItemsTags( const PimItem::List &items, const Tag::List &ta
     return false;
   }
 
-  mNotificationCollector->itemsTagsChanged( items, QSet<qint64>(), removedTags );
+  if ( qb.query().numRowsAffected() != 0 ) {
+    setBoolPtr( tagsChanged, true );
+    if ( !silent ) {
+      mNotificationCollector->itemsTagsChanged( items, QSet<qint64>(), removedTags );
+    }
+  }
+
   return true;
 }
 
@@ -812,12 +869,9 @@ void DataStore::activeCachePolicy( Collection &col )
 
 QVector<Collection> DataStore::virtualCollections( const PimItem &item )
 {
-  QueryBuilder qb( CollectionPimItemRelation::tableName(), QueryBuilder::Select );
+  SelectQueryBuilder<Collection> qb;
   qb.addJoin( QueryBuilder::InnerJoin, Collection::tableName(),
               Collection::idFullColumnName(), CollectionPimItemRelation::leftFullColumnName() );
-  Q_FOREACH ( const QString &columnName, Collection::fullColumnNames() ) {
-    qb.addColumn( columnName );
-  }
   qb.addValueCondition( CollectionPimItemRelation::rightFullColumnName(), Query::Equals, item.id() );
 
   if ( !qb.exec() ) {
@@ -826,10 +880,10 @@ QVector<Collection> DataStore::virtualCollections( const PimItem &item )
     return QVector<Collection>();
   }
 
-  return Collection::extractResult( qb.query() );
+  return qb.result();
 }
 
-QMap<Entity::Id,PimItem> DataStore::virtualCollections( const PimItem::List &items )
+QMap<Entity::Id,QList<PimItem> > DataStore::virtualCollections( const PimItem::List &items )
 {
   QueryBuilder qb( CollectionPimItemRelation::tableName(), QueryBuilder::Select );
   qb.addJoin( QueryBuilder::InnerJoin, Collection::tableName(),
@@ -839,29 +893,42 @@ QMap<Entity::Id,PimItem> DataStore::virtualCollections( const PimItem::List &ite
   qb.addColumn( Collection::idFullColumnName() );
   qb.addColumns( QStringList() << PimItem::idFullColumnName()
                                << PimItem::remoteIdFullColumnName()
-                               << PimItem::remoteRevisionFullColumnName() );
+                               << PimItem::remoteRevisionFullColumnName()
+                               << PimItem::mimeTypeIdFullColumnName() );
+  qb.addSortColumn( Collection::idFullColumnName(), Query::Ascending );
 
-  Query::Condition condition( Query::Or );
-  QStringList ids;
-  Q_FOREACH ( const PimItem &item, items ) {
-    ids << QString::number( item.id() );
+  if ( items.count() == 1) {
+    qb.addValueCondition( CollectionPimItemRelation::rightFullColumnName(), Query::Equals, items.first().id() );
+  } else {
+    QVariantList ids;
+    ids.reserve(items.count());
+    Q_FOREACH ( const PimItem &item, items ) {
+      ids << item.id();
+    }
+    qb.addValueCondition( CollectionPimItemRelation::rightFullColumnName(), Query::In, ids );
   }
-  qb.addValueCondition( CollectionPimItemRelation::rightFullColumnName(), Query::In, ids );
 
   if ( !qb.exec() ) {
     akDebug() << "Error during selection of records from table CollectionPimItemRelation"
               << qb.query().lastError().text();
-    return QMap<Entity::Id,PimItem>();
+    return QMap<Entity::Id, QList<PimItem> >();
   }
 
   QSqlQuery query = qb.query();
-  QMap<Entity::Id,PimItem> map;
-  while ( query.next() ) {
-    PimItem item;
-    item.setId( query.value( 1 ).toLongLong() );
-    item.setRemoteId( query.value( 2 ).toString() );
-    item.setRemoteRevision( query.value( 3 ).toString() );
-    map.insertMulti( query.value( 0 ).toLongLong(), item );
+  QMap<Entity::Id, QList<PimItem> > map;
+  QList<PimItem> pimItems;
+  query.next();
+  while ( query.isValid() ) {
+    const qlonglong collectionId = query.value(0).toLongLong();
+    QList<PimItem> &pimItems = map[collectionId];
+    do {
+      PimItem item;
+      item.setId( query.value( 1 ).toLongLong() );
+      item.setRemoteId( query.value( 2 ).toString() );
+      item.setRemoteRevision( query.value( 3 ).toString() );
+      item.setMimeTypeId( query.value( 4 ).toLongLong() );
+      pimItems << item;
+    } while (query.next() && query.value(0).toLongLong() > collectionId);
   }
 
   return map;
@@ -1083,23 +1150,27 @@ QDateTime DataStore::dateTimeToQDateTime( const QByteArray &dateTime )
 
 void DataStore::addQueryToTransaction( const QSqlQuery &query, bool isBatch )
 {
-  DbType::Type dbType = DbType::type( m_database );
   // This is used for replaying deadlocked transactions, so only record queries
   // for backends that support concurrent transactions.
-  if ( !inTransaction() || ( dbType != DbType::MySQL && dbType != DbType::PostgreSQL ) ) {
+  if ( !inTransaction() || DbType::isSystemSQLite( m_database ) ) {
     return;
   }
 
   m_transactionQueries.append( qMakePair( query, isBatch ) );
 }
 
-QSqlQuery DataStore::retryLastTransaction()
+QSqlQuery DataStore::retryLastTransaction( bool rollbackFirst )
 {
-  DbType::Type dbType = DbType::type( m_database );
-  if ( !inTransaction() || ( dbType != DbType::MySQL && dbType != DbType::PostgreSQL ) ) {
+  if ( !inTransaction() || DbType::isSystemSQLite( m_database ) ) {
     return QSqlQuery();
   }
 
+  if ( rollbackFirst ) {
+    // In some cases the SQL database won't rollback the failed transaction, so
+    // we need to do it manually
+    m_database.driver()->rollbackTransaction();
+  }
+
   // The database has rolled back the actual transaction, so reset the counter
   // to 0 and start a new one in beginTransaction(). Then restore the level
   // because this has to be completely transparent to the original caller
